from qiskit import QuantumCircuit
from qiskit.quantum_info import Clifford, random_clifford

# ---------------------------------------------------------------------
# Coupling maps and permutations
# ---------------------------------------------------------------------

coupling_maps = {
    "3qL": [[0, 1], [1, 0], [1, 2], [2, 1]],
    "4qL": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
    ],
    "4qT": [[0, 1], [1, 0], [1, 2], [1, 3], [2, 1], [3, 1]],
    "5qT": [[0, 1], [1, 0], [1, 2], [1, 3], [2, 1], [3, 1], [3, 4], [4, 3]],
    "5qL": [[0, 1], [1, 0], [1, 2], [2, 1], [2, 3], [3, 2], [3, 4], [4, 3]],
    "6qT": [
        [0, 1],
        [1, 0],
        [1, 2],
        [1, 3],
        [2, 1],
        [3, 1],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
    ],
    "6qY": [
        [0, 1],
        [1, 0],
        [1, 3],
        [3, 1],
        [3, 2],
        [2, 3],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
    ],
    "6qL": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
    ],
    "7qT": [
        [0, 1],
        [1, 0],
        [1, 2],
        [1, 3],
        [2, 1],
        [3, 1],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
    ],
    "7qT2": [
        [0, 1],
        [1, 0],
        [1, 3],
        [3, 1],
        [3, 2],
        [2, 3],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
    ],
    "7qY": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [2, 5],
        [5, 2],
        [5, 6],
        [6, 5],
    ],
    "7qL": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
    ],
    "7qH": [
        [0, 1],
        [1, 0],
        [1, 2],
        [1, 3],
        [2, 1],
        [3, 1],
        [3, 5],
        [4, 5],
        [5, 3],
        [5, 4],
        [5, 6],
        [6, 5],
    ],
    "8qT": [
        [0, 1],
        [1, 0],
        [1, 2],
        [1, 3],
        [2, 1],
        [3, 1],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
    ],
    "8qT2": [
        [0, 1],
        [1, 0],
        [1, 3],
        [3, 1],
        [3, 2],
        [2, 3],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
    ],
    "8qY": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 4],
        [4, 2],
        [4, 7],
        [7, 4],
        [4, 5],
        [5, 4],
        [6, 5],
        [5, 6],
        [3, 6],
        [6, 3],
    ],
    "8qY2": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [3, 6],
        [6, 3],
        [4, 5],
        [5, 4],
        [6, 7],
        [7, 6],
    ],
    "8qL": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
    ],
    "8qH": [
        [0, 2],
        [2, 0],
        [2, 1],
        [1, 2],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [4, 6],
        [6, 4],
        [6, 7],
        [7, 6],
    ],
    "9qT": [
        [0, 1],
        [1, 0],
        [1, 2],
        [1, 3],
        [2, 1],
        [3, 1],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
        [7, 8],
        [8, 7],
    ],
    "9qT2": [
        [0, 1],
        [1, 0],
        [1, 3],
        [3, 1],
        [3, 5],
        [5, 3],
        [2, 4],
        [4, 2],
        [4, 7],
        [7, 4],
        [5, 6],
        [6, 5],
        [5, 8],
        [8, 5],
        [7, 6],
        [6, 7],
    ],
    "9qT3": [
        [0, 1],
        [1, 0],
        [1, 3],
        [3, 1],
        [3, 5],
        [5, 3],
        [2, 4],
        [4, 2],
        [4, 7],
        [7, 4],
        [5, 6],
        [6, 5],
        [7, 6],
        [6, 7],
        [7, 8],
        [8, 7],
    ],
    "9qY": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 4],
        [4, 2],
        [4, 5],
        [5, 4],
        [4, 7],
        [7, 4],
        [3, 6],
        [6, 3],
        [6, 5],
        [5, 6],
        [7, 8],
        [8, 7],
    ],
    "9qY2": [
        [0, 2],
        [2, 0],
        [2, 4],
        [4, 2],
        [1, 3],
        [3, 1],
        [3, 6],
        [6, 3],
        [4, 5],
        [5, 4],
        [4, 7],
        [7, 4],
        [6, 5],
        [5, 6],
        [7, 8],
        [8, 7],
    ],
    "9qL": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
        [7, 8],
        [8, 7],
    ],
    "9qH": [
        [0, 2],
        [2, 0],
        [2, 1],
        [1, 2],
        [2, 3],
        [3, 2],
        [3, 5],
        [5, 3],
        [5, 6],
        [6, 5],
        [5, 8],
        [8, 5],
        [4, 7],
        [7, 4],
        [7, 6],
        [6, 7],
    ],
    "9qH2": [
        [0, 2],
        [2, 0],
        [2, 1],
        [1, 2],
        [2, 3],
        [3, 2],
        [3, 5],
        [5, 3],
        [5, 6],
        [6, 5],
        [4, 7],
        [7, 4],
        [7, 6],
        [6, 7],
        [7, 8],
        [8, 7],
    ],
    "9qH3": [
        [0, 1],
        [1, 3],
        [3, 2],
        [3, 4],
        [4, 5],
        [5, 6],
        [5, 7],
        [7, 8],
        [1, 0],
        [3, 1],
        [2, 3],
        [4, 3],
        [5, 4],
        [6, 5],
        [7, 5],
        [8, 7],
    ],
    "9qH4": [
        [0, 2],
        [2, 0],
        [2, 1],
        [1, 2],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [4, 7],
        [7, 4],
        [5, 6],
        [6, 5],
        [7, 8],
        [8, 7],
    ],
    "10qL": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
        [7, 8],
        [8, 7],
        [8, 9],
        [9, 8],
    ],
    "11qL": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
        [7, 8],
        [8, 7],
        [8, 9],
        [9, 8],
        [9, 10],
        [10, 9],
    ],
    "12qL": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
        [7, 8],
        [8, 7],
        [8, 9],
        [9, 8],
        [9, 10],
        [10, 9],
        [10, 11],
        [11, 10],
    ],
    "12qO": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
        [7, 8],
        [8, 7],
        [8, 9],
        [9, 8],
        [9, 10],
        [10, 9],
        [10, 11],
        [11, 10],
        [11, 0],
        [0, 11],
    ],
    "16qX": [
        [12, 15],
        [15, 12],
        [12, 10],
        [10, 12],
        [7, 4],
        [4, 7],
        [8, 9],
        [9, 8],
        [2, 3],
        [3, 2],
        [5, 3],
        [3, 5],
        [12, 13],
        [13, 12],
        [0, 1],
        [1, 0],
        [11, 8],
        [8, 11],
        [7, 6],
        [6, 7],
        [11, 14],
        [14, 11],
        [5, 8],
        [8, 5],
        [10, 7],
        [7, 10],
        [4, 1],
        [1, 4],
        [2, 1],
        [1, 2],
        [13, 14],
        [14, 13],
    ],
}

valid_perms_topology = {
    "3qL": [[0, 1, 2], [2, 1, 0]],
    "4qL": [[0, 1, 2, 3], [3, 2, 1, 0]],
    "4qT": [
        [0, 1, 2, 3],
        [0, 1, 3, 2],
        [2, 1, 0, 3],
        [2, 1, 3, 0],
        [3, 1, 2, 0],
        [3, 1, 0, 2],
    ],
    "5qT": [[0, 1, 2, 3, 4], [2, 1, 0, 3, 4]],
    "5qL": [[0, 1, 2, 3, 4], [4, 3, 2, 1, 0]],
    "6qT": [[0, 1, 2, 3, 4, 5], [2, 1, 0, 3, 4, 5]],
    "6qY": [[0, 1, 2, 3, 4, 5], [5, 4, 2, 3, 1, 0]],
    "6qL": [[0, 1, 2, 3, 4, 5], [5, 4, 3, 2, 1, 0]],
    "7qT": [[0, 1, 2, 3, 4, 5, 6], [2, 1, 0, 3, 4, 5, 6]],
    "7qT2": [[0, 1, 2, 3, 4, 5, 6]],
    "7qY": [
        [0, 1, 2, 3, 4, 5, 6],
        [0, 1, 2, 5, 6, 3, 4],
        [4, 3, 2, 1, 0, 5, 6],
        [4, 3, 2, 5, 6, 1, 0],
        [6, 5, 2, 1, 0, 3, 4],
        [6, 5, 2, 3, 4, 1, 0],
    ],
    "7qL": [[0, 1, 2, 3, 4, 5, 6], [6, 5, 4, 3, 2, 1, 0]],
    "7qH": [
        [0, 1, 2, 3, 4, 5, 6],
        [0, 1, 2, 3, 6, 5, 4],
        [2, 1, 0, 3, 4, 5, 6],
        [2, 1, 0, 3, 6, 5, 4],
        [4, 5, 6, 3, 0, 1, 2],
        [4, 5, 6, 3, 2, 1, 0],
        [6, 5, 4, 3, 2, 1, 0],
        [6, 5, 4, 3, 0, 1, 2],
    ],
    "8qT": [[0, 1, 2, 3, 4, 5, 6, 7], [2, 1, 0, 3, 4, 5, 6, 7]],
    "8qT2": [[0, 1, 2, 3, 4, 5, 6, 7]],
    "8qY": [[0, 1, 2, 3, 4, 5, 6, 7], [3, 6, 5, 0, 4, 2, 1, 7]],
    "8qY2": [[0, 1, 2, 3, 4, 5, 6, 7], [0, 1, 2, 3, 6, 7, 4, 5]],
    "8qL": [[0, 1, 2, 3, 4, 5, 6, 7], [7, 6, 5, 4, 3, 2, 1, 0]],
    "8qH": [[0, 1, 2, 3, 4, 5, 6, 7], [1, 0, 2, 3, 4, 5, 6, 7]],
    "9qT": [[0, 1, 2, 3, 4, 5, 6, 7, 8], [2, 1, 0, 3, 4, 5, 6, 7, 8]],
    "9qT2": [[0, 1, 2, 3, 4, 5, 6, 7, 8]],
    "9qT3": [[0, 1, 2, 3, 4, 5, 6, 7, 8]],
    "9qY": [[0, 1, 2, 3, 4, 5, 6, 7, 8], [3, 6, 5, 0, 4, 2, 1, 7, 8]],
    "9qY2": [[0, 1, 2, 3, 4, 5, 6, 7, 8], [8, 1, 7, 3, 4, 5, 6, 2, 0]],
    "9qL": [[0, 1, 2, 3, 4, 5, 6, 7, 8], [8, 7, 6, 5, 4, 3, 2, 1, 0]],
    "9qH": [[0, 1, 2, 3, 4, 5, 6, 7, 8], [1, 0, 2, 3, 4, 5, 6, 7, 8]],
    "9qH2": [
        [0, 1, 2, 3, 4, 5, 6, 7, 8],
        [0, 1, 2, 3, 8, 5, 6, 7, 4],
        [1, 0, 2, 3, 4, 5, 6, 7, 8],
        [1, 0, 2, 3, 8, 5, 6, 7, 4],
        [4, 8, 7, 6, 0, 5, 3, 2, 1],
        [4, 8, 7, 6, 1, 5, 3, 2, 0],
        [8, 4, 7, 6, 0, 5, 3, 2, 1],
        [8, 4, 7, 6, 1, 5, 3, 2, 0],
    ],
    "9qH3": [[0, 1, 2, 3, 4, 5, 6, 7, 8], [8, 7, 6, 5, 4, 3, 2, 1, 0]],
    "9qH4": [
        [0, 1, 2, 3, 4, 5, 6, 7, 8],
        [0, 1, 2, 3, 4, 7, 8, 5, 6],
        [1, 0, 2, 3, 4, 5, 6, 7, 8],
        [1, 0, 2, 3, 4, 7, 8, 5, 6],
    ],
    "10qL": [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]],
    "11qL": [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10], [10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0]],
    "12qL": [
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
        [11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0],
    ],
    "12qO": [
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
        [0, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1],
        [1, 0, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2],
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 0],
        [2, 1, 0, 11, 10, 9, 8, 7, 6, 5, 4, 3],
        [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 0, 1],
        [3, 2, 1, 0, 11, 10, 9, 8, 7, 6, 5, 4],
        [3, 4, 5, 6, 7, 8, 9, 10, 11, 0, 1, 2],
        [4, 3, 2, 1, 0, 11, 10, 9, 8, 7, 6, 5],
        [4, 5, 6, 7, 8, 9, 10, 11, 0, 1, 2, 3],
        [5, 4, 3, 2, 1, 0, 11, 10, 9, 8, 7, 6],
        [5, 6, 7, 8, 9, 10, 11, 0, 1, 2, 3, 4],
        [6, 5, 4, 3, 2, 1, 0, 11, 10, 9, 8, 7],
        [6, 7, 8, 9, 10, 11, 0, 1, 2, 3, 4, 5],
        [7, 6, 5, 4, 3, 2, 1, 0, 11, 10, 9, 8],
        [7, 8, 9, 10, 11, 0, 1, 2, 3, 4, 5, 6],
        [8, 7, 6, 5, 4, 3, 2, 1, 0, 11, 10, 9],
        [8, 9, 10, 11, 0, 1, 2, 3, 4, 5, 6, 7],
        [9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 11, 10],
        [9, 10, 11, 0, 1, 2, 3, 4, 5, 6, 7, 8],
        [10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 11],
        [10, 11, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        [11, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        [11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0],
    ],
    "16qX": [
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15],
        [15, 12, 13, 14, 10, 11, 6, 7, 8, 9, 4, 5, 1, 2, 3, 0],
    ],
}


CLIFFORD_COUPLING_MAPS_BY_HASHES_DICT = {
    "5144181ac27497fdfa9bdb5b8b799630": [[0, 1], [1, 0], [1, 2], [2, 1]],
    "69144809aea48cb46eae9c3950f24a15": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
    ],
    "cde6b48ed870286595c1455af7aff8bd": [
        [0, 1],
        [1, 0],
        [1, 2],
        [1, 3],
        [2, 1],
        [3, 1],
    ],
    "7e799239f8573f120a2c5e15cda2e88f": [
        [0, 1],
        [1, 0],
        [1, 2],
        [1, 3],
        [2, 1],
        [3, 1],
        [3, 4],
        [4, 3],
    ],
    "5a8eac0760a558d4174437be478ec0aa": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
    ],
    "f619afa25c1921f996d128a1dfefef72": [
        [0, 1],
        [1, 0],
        [1, 2],
        [1, 3],
        [2, 1],
        [3, 1],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
    ],
    "86f7cf811774c9dc1f8ac7ebefafd51c": [
        [0, 1],
        [1, 0],
        [1, 3],
        [3, 1],
        [3, 2],
        [2, 3],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
    ],
    "50803b752054f0512687e537ee7368f9": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
    ],
    "71e5bb03712e194188f8d50f4c05cdf8": [
        [0, 1],
        [1, 0],
        [1, 2],
        [1, 3],
        [2, 1],
        [3, 1],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
    ],
    "155388a51f67e78b5cb4ef00a89a070e": [
        [0, 1],
        [1, 0],
        [1, 3],
        [3, 1],
        [3, 2],
        [2, 3],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
    ],
    "61f645001e86ad8a32357cc828ae33cb": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [2, 5],
        [5, 2],
        [5, 6],
        [6, 5],
    ],
    "bd4d5e09460e489b3a7687d2fb06fc0c": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
    ],
    "af7ff602e385e2b324d14d3396a2867f": [
        [0, 1],
        [1, 0],
        [1, 2],
        [1, 3],
        [2, 1],
        [3, 1],
        [3, 5],
        [4, 5],
        [5, 3],
        [5, 4],
        [5, 6],
        [6, 5],
    ],
    "ebab7e6f2fee3938288afdd9c22c7514": [
        [0, 1],
        [1, 0],
        [1, 2],
        [1, 3],
        [2, 1],
        [3, 1],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
    ],
    "e1349747c18d9855ba9cdb574ac00adc": [
        [0, 1],
        [1, 0],
        [1, 3],
        [3, 1],
        [3, 2],
        [2, 3],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
    ],
    "d0b1e042c9dd64acbc37de926fe39063": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 4],
        [4, 2],
        [4, 7],
        [7, 4],
        [4, 5],
        [5, 4],
        [6, 5],
        [5, 6],
        [3, 6],
        [6, 3],
    ],
    "ef72178b7a1a1759663885bdd083eae0": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [3, 6],
        [6, 3],
        [4, 5],
        [5, 4],
        [6, 7],
        [7, 6],
    ],
    "653d67ad3d369bf2cd63d0027cde92ec": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
    ],
    "2e7001dcd306499fb6e11ddc4eac9068": [
        [0, 2],
        [2, 0],
        [2, 1],
        [1, 2],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [4, 6],
        [6, 4],
        [6, 7],
        [7, 6],
    ],
    "187ebadd19b16fcd271a422da011396d": [
        [0, 1],
        [1, 0],
        [1, 2],
        [1, 3],
        [2, 1],
        [3, 1],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
        [7, 8],
        [8, 7],
    ],
    # TODO: Review if this should exists on the config.json for clifford
    # "9qT2": [
    #     [0, 1],
    #     [1, 0],
    #     [1, 3],
    #     [3, 1],
    #     [3, 5],
    #     [5, 3],
    #     [2, 4],
    #     [4, 2],
    #     [4, 7],
    #     [7, 4],
    #     [5, 6],
    #     [6, 5],
    #     [5, 8],
    #     [8, 5],
    #     [7, 6],
    #     [6, 7],
    # ],
    # "9qT3": [
    #     [0, 1],
    #     [1, 0],
    #     [1, 3],
    #     [3, 1],
    #     [3, 5],
    #     [5, 3],
    #     [2, 4],
    #     [4, 2],
    #     [4, 7],
    #     [7, 4],
    #     [5, 6],
    #     [6, 5],
    #     [7, 6],
    #     [6, 7],
    #     [7, 8],
    #     [8, 7],
    # ],
    "b037f8c76e9f028841208ecb7d8f4737": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 4],
        [4, 2],
        [4, 5],
        [5, 4],
        [4, 7],
        [7, 4],
        [3, 6],
        [6, 3],
        [6, 5],
        [5, 6],
        [7, 8],
        [8, 7],
    ],
    "4703b05fe1aa4407e34296678b8ac8df": [
        [0, 2],
        [2, 0],
        [2, 4],
        [4, 2],
        [1, 3],
        [3, 1],
        [3, 6],
        [6, 3],
        [4, 5],
        [5, 4],
        [4, 7],
        [7, 4],
        [6, 5],
        [5, 6],
        [7, 8],
        [8, 7],
    ],
    "44003dd512e49164eb3c5d08bf21eb36": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
        [7, 8],
        [8, 7],
    ],
    "ff7241d51cdb655d6dec2f945af83fc0": [
        [0, 2],
        [2, 0],
        [2, 1],
        [1, 2],
        [2, 3],
        [3, 2],
        [3, 5],
        [5, 3],
        [5, 6],
        [6, 5],
        [5, 8],
        [8, 5],
        [4, 7],
        [7, 4],
        [7, 6],
        [6, 7],
    ],
    "c4a271fff109309e7f31bc6df27c4553": [
        [0, 2],
        [2, 0],
        [2, 1],
        [1, 2],
        [2, 3],
        [3, 2],
        [3, 5],
        [5, 3],
        [5, 6],
        [6, 5],
        [4, 7],
        [7, 4],
        [7, 6],
        [6, 7],
        [7, 8],
        [8, 7],
    ],
    "aba279ad6cfa269cbd11f404fda2b8f0": [
        [0, 1],
        [1, 3],
        [3, 2],
        [3, 4],
        [4, 5],
        [5, 6],
        [5, 7],
        [7, 8],
        [1, 0],
        [3, 1],
        [2, 3],
        [4, 3],
        [5, 4],
        [6, 5],
        [7, 5],
        [8, 7],
    ],
    "80e7f02f768b5a7a2366304648dc596a": [
        [0, 2],
        [2, 0],
        [2, 1],
        [1, 2],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [4, 7],
        [7, 4],
        [5, 6],
        [6, 5],
        [7, 8],
        [8, 7],
    ],
    "dc09eb5accc563a82d933e0dde59da6e": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
        [7, 8],
        [8, 7],
        [8, 9],
        [9, 8],
    ],
    "eac746408de2dfc79940e1d9fa12da5e": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
        [7, 8],
        [8, 7],
        [8, 9],
        [9, 8],
        [9, 10],
        [10, 9],
    ],
    # TODO: Review if this should exists on the config.json for clifford
    # "12qL": [
    #     [0, 1],
    #     [1, 0],
    #     [1, 2],
    #     [2, 1],
    #     [2, 3],
    #     [3, 2],
    #     [3, 4],
    #     [4, 3],
    #     [4, 5],
    #     [5, 4],
    #     [5, 6],
    #     [6, 5],
    #     [6, 7],
    #     [7, 6],
    #     [7, 8],
    #     [8, 7],
    #     [8, 9],
    #     [9, 8],
    #     [9, 10],
    #     [10, 9],
    #     [10, 11],
    #     [11, 10],
    # ],
    "b8a6d35c01a1367bb3145d687fac2c5d": [
        [0, 1],
        [1, 0],
        [1, 2],
        [2, 1],
        [2, 3],
        [3, 2],
        [3, 4],
        [4, 3],
        [4, 5],
        [5, 4],
        [5, 6],
        [6, 5],
        [6, 7],
        [7, 6],
        [7, 8],
        [8, 7],
        [8, 9],
        [9, 8],
        [9, 10],
        [10, 9],
        [10, 11],
        [11, 10],
        [11, 0],
        [0, 11],
    ],
    # TODO: Review if this should exists on the config.json for clifford
    # "16qX": [
    #     [12, 15],
    #     [15, 12],
    #     [12, 10],
    #     [10, 12],
    #     [7, 4],
    #     [4, 7],
    #     [8, 9],
    #     [9, 8],
    #     [2, 3],
    #     [3, 2],
    #     [5, 3],
    #     [3, 5],
    #     [12, 13],
    #     [13, 12],
    #     [0, 1],
    #     [1, 0],
    #     [11, 8],
    #     [8, 11],
    #     [7, 6],
    #     [6, 7],
    #     [11, 14],
    #     [14, 11],
    #     [5, 8],
    #     [8, 5],
    #     [10, 7],
    #     [7, 10],
    #     [4, 1],
    #     [1, 4],
    #     [2, 1],
    #     [1, 2],
    #     [13, 14],
    #     [14, 13],
    # ],
}


def embed_clifford(cliff, nq):
    new_cliff = Clifford(QuantumCircuit(nq))
    oq = cliff.num_qubits
    new_cliff.stab_x[:oq, :oq] = cliff.stab_x[:, :]
    new_cliff.stab_z[:oq, :oq] = cliff.stab_z[:, :]
    new_cliff.stab_phase[:oq] = cliff.stab_phase[:]

    new_cliff.destab_x[:oq, :oq] = cliff.destab_x[:, :]
    new_cliff.destab_z[:oq, :oq] = cliff.destab_z[:, :]
    new_cliff.destab_phase[:oq] = cliff.destab_phase[:]

    return new_cliff


def get_random_clifford_operator(num_qubits, seed=42) -> Clifford:
    return random_clifford(num_qubits, seed=seed)


def check_synthesized_clifford(
    original_clifford: Clifford, synthesized_qc: QuantumCircuit
):
    """Check whether a synthesized circuit does the same as another original clifford"""
    synthesized_clifford = Clifford(synthesized_qc)
    if original_clifford.num_qubits != synthesized_clifford.num_qubits:
        # We can have the situation in transpiling that the synthesized clifford needs to use
        # more qubits than the original Clifford because of the topology and its shape
        # In that case, we should embed the smaller Clifford in the max N of qubits before comparing both
        max_n_qubits = max(
            original_clifford.num_qubits, synthesized_clifford.num_qubits
        )
        if original_clifford.num_qubits > synthesized_clifford.num_qubits:
            synthesized_clifford = embed_clifford(
                cliff=synthesized_clifford, nq=max_n_qubits
            )
        else:
            original_clifford = embed_clifford(cliff=original_clifford, nq=max_n_qubits)
    return synthesized_clifford == original_clifford
