# -*- coding: utf-8 -*-

# (C) Copyright 2023 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Utils for Linear module"""

from collections import OrderedDict

from qiskit.transpiler.passes.routing import SabreSwap
from qiskit.transpiler import CouplingMap
from qiskit.circuit.library.generalized_gates.linear_function import LinearFunction
from qiskit import QuantumCircuit

import numpy as np
from qiskit import QuantumCircuit
from qiskit.circuit.library.generalized_gates.linear_function import LinearFunction
from qiskit.converters import circuit_to_dag, dag_to_circuit
from qiskit.quantum_info import Clifford
from qiskit.synthesis.linear.linear_matrix_utils import random_invertible_binary_matrix

import networkx as nx
import logging


logger = logging.getLogger(__name__)

MODEL_NAMES = [
    "3qL",
    "4qL",
    "4qY",
    "5qL",
    "5qT",
    "6qL",
    "6qT",
    "6qY",
    "7qF",
    "7qH",
    "7qL",
    "7qT",
    "7qY",
    "8qF",
    "8qJ",
    "8qL",
    "8qT1",
    "8qT2",
    "8qY",
    "9qF1",
    "9qF2",
    "9qH1",
    "9qH2",
    "9qH3",
    "9qJ",
    "9qL",
    "9qT1",
    "9qT2",
    "9qY",
    "2q",
]
MODEL_QUBITS = [int(name[0]) for name in MODEL_NAMES]
MODEL_HASHES = [
    "5144181ac27497fdfa9bdb5b8b799630",  # pragma: allowlist-secret
    "69144809aea48cb46eae9c3950f24a15",  # pragma: allowlist-secret
    "cde6b48ed870286595c1455af7aff8bd",  # pragma: allowlist-secret
    "5a8eac0760a558d4174437be478ec0aa",  # pragma: allowlist-secret
    "7e799239f8573f120a2c5e15cda2e88f",  # pragma: allowlist-secret
    "50803b752054f0512687e537ee7368f9",  # pragma: allowlist-secret
    "f619afa25c1921f996d128a1dfefef72",  # pragma: allowlist-secret
    "86f7cf811774c9dc1f8ac7ebefafd51c",  # pragma: allowlist-secret
    "155388a51f67e78b5cb4ef00a89a070e",  # pragma: allowlist-secret
    "af7ff602e385e2b324d14d3396a2867f",  # pragma: allowlist-secret
    "bd4d5e09460e489b3a7687d2fb06fc0c",  # pragma: allowlist-secret
    "71e5bb03712e194188f8d50f4c05cdf8",  # pragma: allowlist-secret
    "61f645001e86ad8a32357cc828ae33cb",  # pragma: allowlist-secret
    "e1349747c18d9855ba9cdb574ac00adc",  # pragma: allowlist-secret
    "2e7001dcd306499fb6e11ddc4eac9068",  # pragma: allowlist-secret
    "653d67ad3d369bf2cd63d0027cde92ec",  # pragma: allowlist-secret
    "ebab7e6f2fee3938288afdd9c22c7514",  # pragma: allowlist-secret
    "ef72178b7a1a1759663885bdd083eae0",  # pragma: allowlist-secret
    "d0b1e042c9dd64acbc37de926fe39063",  # pragma: allowlist-secret
    "e0686f2bdd1b5066f9bc4ad94630b16b",  # pragma: allowlist-secret
    "47156d2004c0e071ee2e10f1613b2f71",  # pragma: allowlist-secret
    "c4a271fff109309e7f31bc6df27c4553",  # pragma: allowlist-secret
    "aba279ad6cfa269cbd11f404fda2b8f0",  # pragma: allowlist-secret
    "80e7f02f768b5a7a2366304648dc596a",  # pragma: allowlist-secret
    "ff7241d51cdb655d6dec2f945af83fc0",  # pragma: allowlist-secret
    "44003dd512e49164eb3c5d08bf21eb36",  # pragma: allowlist-secret
    "187ebadd19b16fcd271a422da011396d",  # pragma: allowlist-secret
    "4703b05fe1aa4407e34296678b8ac8df",  # pragma: allowlist-secret
    "b037f8c76e9f028841208ecb7d8f4737",  # pragma: allowlist-secret
    "3ba4ffe16dfe637510ed1c3676ec6cb0",  # pragma: allowlist-secret # EXTRA 2q GRAPH
]

MODEL_CMAPS = [
    [(0, 2), (1, 2), (2, 0), (2, 1)],
    [(0, 1), (0, 3), (1, 0), (2, 3), (3, 0), (3, 2)],
    [(0, 3), (1, 3), (2, 3), (3, 0), (3, 1), (3, 2)],
    [(0, 1), (1, 0), (1, 2), (2, 1), (2, 3), (3, 2), (3, 4), (4, 3)],
    [(0, 2), (1, 2), (2, 0), (2, 1), (2, 3), (3, 2), (3, 4), (4, 3)],
    [(0, 1), (1, 0), (1, 2), (2, 1), (2, 3), (3, 2), (3, 4), (4, 3), (4, 5), (5, 4)],
    [(0, 2), (1, 2), (2, 0), (2, 1), (2, 3), (3, 2), (3, 4), (4, 3), (4, 5), (5, 4)],
    [(0, 1), (1, 0), (1, 3), (2, 3), (3, 1), (3, 2), (3, 4), (4, 3), (4, 5), (5, 4)],
    [
        (0, 1),
        (1, 0),
        (1, 3),
        (2, 3),
        (3, 1),
        (3, 2),
        (3, 4),
        (4, 3),
        (4, 5),
        (5, 4),
        (5, 6),
        (6, 5),
    ],
    [
        (0, 2),
        (1, 2),
        (2, 0),
        (2, 1),
        (2, 3),
        (3, 2),
        (3, 4),
        (4, 3),
        (4, 5),
        (4, 6),
        (5, 4),
        (6, 4),
    ],
    [
        (0, 1),
        (1, 0),
        (1, 2),
        (2, 1),
        (2, 4),
        (3, 6),
        (4, 2),
        (4, 5),
        (5, 4),
        (5, 6),
        (6, 3),
        (6, 5),
    ],
    [
        (0, 2),
        (1, 2),
        (2, 0),
        (2, 1),
        (2, 3),
        (3, 2),
        (3, 4),
        (4, 3),
        (4, 5),
        (5, 4),
        (5, 6),
        (6, 5),
    ],
    [
        (0, 1),
        (1, 0),
        (1, 2),
        (2, 1),
        (2, 3),
        (2, 5),
        (3, 2),
        (3, 4),
        (4, 3),
        (5, 2),
        (5, 6),
        (6, 5),
    ],
    [
        (0, 1),
        (1, 0),
        (1, 3),
        (2, 3),
        (3, 1),
        (3, 2),
        (3, 4),
        (4, 3),
        (4, 5),
        (5, 4),
        (5, 6),
        (6, 5),
        (6, 7),
        (7, 6),
    ],
    [
        (0, 2),
        (1, 2),
        (2, 0),
        (2, 1),
        (2, 3),
        (3, 2),
        (3, 4),
        (4, 3),
        (4, 5),
        (4, 6),
        (5, 4),
        (6, 4),
        (6, 7),
        (7, 6),
    ],
    [
        (0, 1),
        (1, 0),
        (1, 3),
        (2, 4),
        (3, 1),
        (3, 5),
        (4, 2),
        (4, 7),
        (5, 3),
        (5, 6),
        (6, 5),
        (6, 7),
        (7, 4),
        (7, 6),
    ],
    [
        (0, 2),
        (1, 2),
        (2, 0),
        (2, 1),
        (2, 3),
        (3, 2),
        (3, 5),
        (4, 7),
        (5, 3),
        (5, 6),
        (6, 5),
        (6, 7),
        (7, 4),
        (7, 6),
    ],
    [
        (0, 1),
        (1, 0),
        (1, 2),
        (2, 1),
        (2, 3),
        (3, 2),
        (3, 4),
        (3, 6),
        (4, 3),
        (4, 5),
        (5, 4),
        (6, 3),
        (6, 7),
        (7, 6),
    ],
    [
        (0, 1),
        (1, 0),
        (1, 2),
        (2, 1),
        (2, 4),
        (3, 6),
        (4, 2),
        (4, 5),
        (4, 7),
        (5, 4),
        (5, 6),
        (6, 3),
        (6, 5),
        (7, 4),
    ],
    [
        (0, 1),
        (1, 0),
        (1, 3),
        (2, 4),
        (3, 1),
        (3, 5),
        (4, 2),
        (4, 7),
        (5, 3),
        (5, 6),
        (6, 5),
        (6, 7),
        (7, 4),
        (7, 6),
        (7, 8),
        (8, 7),
    ],
    [
        (0, 1),
        (1, 0),
        (1, 3),
        (2, 4),
        (3, 1),
        (3, 5),
        (4, 2),
        (4, 7),
        (5, 3),
        (5, 6),
        (5, 8),
        (6, 5),
        (6, 7),
        (7, 4),
        (7, 6),
        (8, 5),
    ],
    [
        (0, 2),
        (1, 2),
        (2, 0),
        (2, 1),
        (2, 3),
        (3, 2),
        (3, 5),
        (4, 7),
        (5, 3),
        (5, 6),
        (6, 5),
        (6, 7),
        (7, 4),
        (7, 6),
        (7, 8),
        (8, 7),
    ],
    [
        (0, 1),
        (1, 0),
        (1, 3),
        (2, 3),
        (3, 1),
        (3, 2),
        (3, 4),
        (4, 3),
        (4, 5),
        (5, 4),
        (5, 6),
        (5, 7),
        (6, 5),
        (7, 5),
        (7, 8),
        (8, 7),
    ],
    [
        (0, 2),
        (1, 2),
        (2, 0),
        (2, 1),
        (2, 3),
        (3, 2),
        (3, 4),
        (4, 3),
        (4, 5),
        (4, 7),
        (5, 4),
        (5, 6),
        (6, 5),
        (7, 4),
        (7, 8),
        (8, 7),
    ],
    [
        (0, 2),
        (1, 2),
        (2, 0),
        (2, 1),
        (2, 3),
        (3, 2),
        (3, 5),
        (4, 7),
        (5, 3),
        (5, 6),
        (5, 8),
        (6, 5),
        (6, 7),
        (7, 4),
        (7, 6),
        (8, 5),
    ],
    [
        (0, 1),
        (1, 0),
        (1, 4),
        (2, 3),
        (2, 5),
        (3, 2),
        (4, 1),
        (4, 6),
        (5, 2),
        (5, 8),
        (6, 4),
        (6, 7),
        (7, 6),
        (7, 8),
        (8, 5),
        (8, 7),
    ],
    [
        (0, 2),
        (1, 2),
        (2, 0),
        (2, 1),
        (2, 4),
        (3, 5),
        (4, 2),
        (4, 6),
        (5, 3),
        (5, 8),
        (6, 4),
        (6, 7),
        (7, 6),
        (7, 8),
        (8, 5),
        (8, 7),
    ],
    [
        (0, 2),
        (1, 3),
        (2, 0),
        (2, 4),
        (3, 1),
        (3, 6),
        (4, 2),
        (4, 5),
        (4, 7),
        (5, 4),
        (5, 6),
        (6, 3),
        (6, 5),
        (7, 4),
        (7, 8),
        (8, 7),
    ],
    [
        (0, 1),
        (1, 0),
        (1, 2),
        (2, 1),
        (2, 4),
        (3, 6),
        (4, 2),
        (4, 5),
        (4, 7),
        (5, 4),
        (5, 6),
        (6, 3),
        (6, 5),
        (7, 4),
        (7, 8),
        (8, 7),
    ],
    [(0, 1), (1, 0)],  # EXTRA 2q GRAPH
]

linear_function_coupling_map_by_hash_dict = {
    coupling_map_hash: coupling_map
    for coupling_map_hash, coupling_map in zip(MODEL_HASHES, MODEL_CMAPS)
}

hash_to_cmap = {h: cm for h, cm in zip(MODEL_HASHES, MODEL_CMAPS)}


def get_mapping_perm(coupling_map: nx.Graph, qargs):
    # We first find the subgraph of the coupling map where the Clifford is
    subgraph = coupling_map.subgraph(qargs)

    # Check if it is connected
    if not nx.is_connected(subgraph):
        logger.warning(f"Subgraph {subgraph} is not connected")
        return None, None, False, None

    # We find which model to use by hashing the input graph
    cmap_hash = nx.weisfeiler_lehman_graph_hash(subgraph)

    # Check if there is a model for that subgraph
    if cmap_hash not in hash_to_cmap:
        logger.warning(f"Subgraph hash {cmap_hash} not in available coupling maps")
        return None, None, True, None
    model_cmap = hash_to_cmap[cmap_hash]
    # We then find a mapping between the subgraph and the model's topology
    cmap_to_model = next(
        iter(nx.isomorphism.GraphMatcher(subgraph, nx.Graph(model_cmap)).match())
    )

    # We now have to find the permutation that we should apply to the Clifford based on the mapping we found
    qargs_dict = {v: i for i, v in enumerate(qargs)}
    subgraph_perm = [
        qargs_dict[v]
        for v in sorted(cmap_to_model.keys(), key=lambda k: cmap_to_model[k])
    ]

    return subgraph_perm, cmap_hash, True, model_cmap


def embed_clifford(cliff, nq):
    new_cliff = Clifford(QuantumCircuit(nq))
    oq = cliff.num_qubits
    new_cliff.stab_x[:oq, :oq] = cliff.stab_x[:, :]
    new_cliff.stab_z[:oq, :oq] = cliff.stab_z[:, :]
    new_cliff.stab_phase[:oq] = cliff.stab_phase[:]

    new_cliff.destab_x[:oq, :oq] = cliff.destab_x[:, :]
    new_cliff.destab_z[:oq, :oq] = cliff.destab_z[:, :]
    new_cliff.destab_phase[:oq] = cliff.destab_phase[:]

    return new_cliff


# From this StackOverflow answer:
# https://quantumcomputing.stackexchange.com/questions/25672/remove-inactive-qubits-from-qiskit-circuit
def remove_idle_qwires(circ):
    dag = circuit_to_dag(circ)

    idle_wires = list(dag.idle_wires())
    for w in idle_wires:
        dag._remove_idle_wire(w)
        dag.qubits.remove(w)

    dag.qregs = OrderedDict()

    temp_circuit = dag_to_circuit(dag)

    result_circuit = QuantumCircuit(temp_circuit.num_qubits)
    result_circuit = result_circuit.compose(temp_circuit)
    return result_circuit


def perm_cliff(cliff, perm):
    perm = np.array(perm)
    cliff.stab_x = cliff.stab_x[:, perm]
    cliff.stab_z = cliff.stab_z[:, perm]
    cliff.destab_x = cliff.destab_x[:, perm]
    cliff.destab_z = cliff.destab_z[:, perm]
    cliff.stab = cliff.stab[perm, :]
    cliff.destab = cliff.destab[perm, :]
    return cliff


def random_linfunc_as_clifford(n_qubits, seed=123):
    """Generate a random clifford from a random linear function of n_qubits qubits."""

    random_linear = lambda seed: LinearFunction(
        random_invertible_binary_matrix(n_qubits, seed=seed)
    )
    random_clifford = Clifford(random_linear(seed))
    return random_clifford


def decompose_linear_qiskit_sabre(cliff, coupling_map):
    """Decompose using Sabre Swap"""
    cmap = CouplingMap(coupling_map)
    cmap.make_symmetric()
    ss = SabreSwap(cmap)

    # Make linear function object
    linear_func = LinearFunction(cliff)

    # Synth circuit
    lf_circuit_synth = linear_func.synthesize()
    lf_circuit_synth = QuantumCircuit(lf_circuit_synth.num_qubits).compose(
        lf_circuit_synth
    )

    # Route circuit
    lf_circuit_routed = ss(lf_circuit_synth)

    return lf_circuit_routed
